/* Acromag AVME944X device support for the Binary Output (bo) record */

/*
 * This is a 'quick and dirty' approach.  There is a lot of device support out
 * there like this.  The base address and the number of cards can be changed,
 * but cards have to be placed on adjacent 1k-byte boundaries.  The card number 
 * in the record's OUT field selects a card from the addressed order.  To 
 * change the base address or number of cards, put lines in the startup script 
 * like this (before the call to iocInit):
 *	a944xBase = 0x4000	address of card 0
 *	a944xCards = 2		number of cards
 *
 * Andrew Johnson <anj@aps.anl.gov>
 */

/* vxWorks headers */
#include <vxWorks.h>
#include <sysLib.h>
#include <vxLib.h>
#include <intLib.h>
#include <stdio.h>
#include <vme.h>

/* EPICS headers */
#include <devSup.h>
#include <recSup.h>
#include <boRecord.h>


/*
 * Local declarations and definitions
 */

/* Default base address and number of cards */

long a944xBase = 0x2800;
int a944xCards = 1;


/* Card register structure */

struct a944x {
    uint16_t idProm[0x20];	/* 00-3f */
    uint16_t pad1[0x20];	/* 40-7f */
    uint16_t boardStatus;	/* 80-81 */
    uint16_t pad2[0x0f];	/* 82-9f */
    uint16_t intVector[8];	/* a0-af */
    uint16_t pad3[8];		/* b0-bf */
    uint16_t intStatus; 	/* c0-c1 */
    uint16_t intEnable; 	/* c2-c3 */
    uint16_t intPolarity;	/* c4-c5 */
    uint16_t intTypeSel;	/* c6-c7 */
    uint16_t intPattern;	/* c8-c9 */
    uint16_t inputData; 	/* ca-cb */
    uint16_t outputData;	/* cc-cd */
    uint16_t pad4[0x199];	/* ce-3ff */
};


/* Status report function */

static long report(int level) {
    printf("  Configured for %d cards starting at A16 address %#lx\n", 
	   a944xCards, a944xBase);
    
    return 0;
}


/* Initialize a record */

static long initRecord(struct boRecord *precord) {
    char *vmeAddr;
    uint16_t leds;
    volatile struct a944x *pcard;
    
    precord->pact = 1;
    
    if ((precord->out.type != VME_IO) || 
	(precord->out.value.vmeio.card < 0) ||
	(precord->out.value.vmeio.card >= a944xCards) ||
	(precord->out.value.vmeio.signal < 0) ||
	(precord->out.value.vmeio.signal > 15)) {
	recGblRecordError(S_dev_badOutType, (void *)precord,
		"devBoAvme944x: Bad OUT address");
	return S_dev_badOutType;
    }
    
    vmeAddr = (char *) a944xBase + precord->out.value.vmeio.card * sizeof(struct a944x);
    if (sysBusToLocalAdrs(VME_AM_SUP_SHORT_IO, vmeAddr, 
		(char **)&pcard) == ERROR) {
	recGblRecordError(S_dev_badCard, (void *)precord,
		"devBoAvme944x: VME address conversion failed");
	return S_dev_badCard;
    }
    
    leds = 0x03;	/* 3 means Red=Off, Green=On */
    if (vxMemProbe((char *)&(pcard->boardStatus), WRITE, 
    		   sizeof(leds), (char *)&leds) < 0) {
	recGblRecordError(S_dev_badCard, (void *)precord,
		"devBoAvme944x: AVME944X card not present");
	return S_dev_badCard;
    }
    
    precord->dpvt = pcard;
    precord->mask = 1 << precord->out.value.vmeio.signal;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


/* Hardware output routine */

static long write(struct boRecord *precord) {
    int lockKey;
    volatile struct a944x *pcard = (struct a944x *) precord->dpvt;
    
    lockKey = intLock();
    pcard->outputData = (pcard->outputData & ~precord->mask) |
    			(precord->rval & precord->mask);
    intUnlock(lockKey);
    
    return 0;
}


/* Device Support Entry Table */

struct boDset {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct boRecord *precord);
    long (*getIoIntInfo)(int cmd, struct boRecord *precord, void *ppvt);
    long (*write)(struct boRecord *precord);
} devBoAvme944x = {
    5,
    report,
    NULL,	/* initialize, */
    initRecord,
    NULL,	/* getIoIntInfo, */
    write
};
