/* Acromag AVME944X device support for Multi-Bit Binary Output (mbbo) record */

/*
 * Another 'quick and dirty' approach.  To avoid having to set the card base 
 * addresses more than once, this assumes that the bo device support is present
 * and just reuses its a944xBase and a944xCards global variables.  It also 
 * doesn't repeat the report function, but these differences make the bo driver 
 * "special" in an unsatisfactory way.
 *
 * Andrew Johnson <anj@aps.anl.gov>
 */

/* vxWorks headers */
#include <vxWorks.h>
#include <sysLib.h>
#include <vxLib.h>
#include <intLib.h>
#include <stdio.h>
#include <vme.h>

/* EPICS headers */
#include <devSup.h>
#include <recSup.h>
#include <mbboRecord.h>


/*
 * Local declarations and definitions
 */

/* Base address and number of cards are provided by the bo device support! */

extern long a944xBase;
extern int a944xCards;


/* Card register structure */

struct a944x {
    uint16_t idProm[0x20];	/* 00-3f */
    uint16_t pad1[0x20];	/* 40-7f */
    uint16_t boardStatus;	/* 80-81 */
    uint16_t pad2[0x0f];	/* 82-9f */
    uint16_t intVector[8];	/* a0-af */
    uint16_t pad3[8];		/* b0-bf */
    uint16_t intStatus; 	/* c0-c1 */
    uint16_t intEnable; 	/* c2-c3 */
    uint16_t intPolarity;	/* c4-c5 */
    uint16_t intTypeSel;	/* c6-c7 */
    uint16_t intPattern;	/* c8-c9 */
    uint16_t inputData; 	/* ca-cb */
    uint16_t outputData;	/* cc-cd */
    uint16_t pad4[0x199];	/* ce-3ff */
};


/* Initialize a record */

static long initRecord(struct mbboRecord *precord) {
    char *vmeAddr;
    uint16_t leds;
    volatile struct a944x *pcard;
    
    precord->pact = 1;
    
    if ((precord->out.type != VME_IO) || 
	(precord->out.value.vmeio.card < 0) ||
	(precord->out.value.vmeio.card >= a944xCards) ||
	(precord->out.value.vmeio.signal < 0) ||
	(precord->out.value.vmeio.signal + precord->nobt > 15)) {
	recGblRecordError(S_dev_badOutType, (void *)precord,
		"devMbboAvme944x: Bad OUT address");
	return S_dev_badOutType;
    }
    
    vmeAddr = (char *) a944xBase + precord->out.value.vmeio.card * sizeof(struct a944x);
    if (sysBusToLocalAdrs(VME_AM_SUP_SHORT_IO, vmeAddr, 
		(char **)&pcard) == ERROR) {
	recGblRecordError(S_dev_badCard, (void *)precord,
		"devMbboAvme944x: VME address conversion failed");
	return S_dev_badCard;
    }
    
    leds = 0x03;	/* 3 means Red=Off, Green=On */
    if (vxMemProbe((char *)&(pcard->boardStatus), WRITE, 
    		   sizeof(leds), (char *)&leds) < 0) {
	recGblRecordError(S_dev_badCard, (void *)precord,
		"devMbboAvme944x: AVME944X card not present");
	return S_dev_badCard;
    }
    
    precord->dpvt = pcard;
    precord->shft = precord->out.value.vmeio.signal;
    precord->mask <<= precord->shft;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


/* Hardware output routine */

static long write(struct mbboRecord *precord) {
    int lockKey;
    volatile struct a944x *pcard = (struct a944x *) precord->dpvt;
    
    lockKey = intLock();
    pcard->outputData = (pcard->outputData & ~precord->mask) |
    			(precord->rval & precord->mask);
    intUnlock(lockKey);
    
    return 0;
}


/* Device Support Entry Table */

struct mbboDset {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct mbboRecord *precord);
    long (*getIoIntInfo)(int cmd, struct mbboRecord *precord, void *ppvt);
    long (*write)(struct mbboRecord *precord);
} devMbboAvme944x = {
    5,
    NULL,
    NULL,	/* initialize, */
    initRecord,
    NULL,	/* getIoIntInfo, */
    write
};
