/***** Acromag ACR9440/9443/9447 device support for standard record types *****
 *
 * Author:	Andrew Johnson <anj@aps.anl.gov>
 * Date:	11 June 1999
 */
 
/* vxWorks headers */
#include <vxWorks.h>
#include <sysLib.h>
#include <vxLib.h>
#include <intLib.h>
#include <vme.h>
#include <iv.h>
#include <stdio.h>
#include <stdlib.h>

/* EPICS headers */
#include <drvSup.h>
#include <devSup.h>
#include <recSup.h>
#include <dbScan.h>
#include <boRecord.h>
#include <biRecord.h>
#include <mbboRecord.h>
#include <mbbiRecord.h>
#include <mbboDirectRecord.h>
#include <mbbiDirectRecord.h>

/*
 * Local declarations and definitions
 */

/* Fixing this makes things a bit simpler, and is unlikely to be a problem */

#define MAX_CARDS	20


/* This is the board register map */

struct acr944x {		/* Offset */
    uint16_t idProm[0x20];	/* 00-3f */
    uint16_t pad1[0x20];	/* 40-7f */
    uint16_t boardStatus;	/* 80-81 */
    uint16_t pad2[0x0f];	/* 82-9f */
    uint16_t intVector[8];	/* a0-af */
    uint16_t pad3[8];		/* b0-bf */
    uint16_t intStatus; 	/* c0-c1 */
    uint16_t intEnable; 	/* c2-c3 */
    uint16_t intPolarity;	/* c4-c5 */
    uint16_t intTypeSel;	/* c6-c7 */
    uint16_t intPattern;	/* c8-c9 */
    uint16_t inputData; 	/* ca-cb */
    uint16_t outputData;	/* cc-cd */
    uint16_t pad4[0x199];	/* ce-3ff */
};

/* Some definitions for the boardStatus Register */

#define BSR_LED_RED	0x01	/* 0=on */
#define BSR_LED_GREEN	0x02	/* 1=on */
#define BSR_INT_PENDING	0x04
#define BSR_INT_ENABLE	0x08
#define BSR_RESET	0x10


/* This is device support private data */

struct dpvt {
    int addrType;
    long cardBase;
    volatile struct acr944x *pcard;
    enum cardType {
	cardIO, 	/* Acr9440 */
	cardOut,	/* Acr9443 */
	cardIn  	/* Acr9447 */
    } cardType;
    int intVec;
    IOSCANPVT ioscan[9];	/* One for each bit plus one for mbbi's */
};


/* Array of pointers to private data */

static struct dpvt *(device[MAX_CARDS]);


/* Interrupt Service Routine (input signals only) */

void intRoutine(int parameter) {
    struct dpvt *pdpvt = (struct dpvt *) parameter;
    volatile struct acr944x *pcard = pdpvt->pcard;
    int intStatus = pcard->intStatus;
    int i, b;
    
    for (i=0, b=1; i<8; i++, b<<=1) {
    	if (intStatus & b) {
	    pcard->intStatus = b;	/* Clear this int channel */
	    if (pdpvt->ioscan[i]) scanIoRequest(pdpvt->ioscan[i]);
	}
    }
    if (pdpvt->ioscan[8]) scanIoRequest(pdpvt->ioscan[8]);
    
    return;
}


/* Register a card address at startup */

long acr944xAdd(int cardNum, int addrType, long cardBase, 
		int intLevel, int intVec) {
    volatile struct acr944x *pcard;
    struct dpvt *pdpvt;
    int am, i;
    enum cardType cardType;
    uint16_t dummy;
    static const char id[] = "VMEIDACR944";
    
    if ((cardNum < 0) || (cardNum >= MAX_CARDS)) {
	epicsPrintf("acr944xAdd: cardNum %d out of range (0 - %d)\n",
		    cardNum, MAX_CARDS-1);
	return -1;
    }
    
    if (device[cardNum]) {
    	epicsPrintf("acr944xAdd: cardNum %d already used (A%d %#x)\n", cardNum,
		    device[cardNum]->addrType, device[cardNum]->cardBase);
	return -1;
    }
    
    switch (addrType) {
    case 16:
    	if ((cardBase < 0) || (cardBase > 0xfc00)) {
	    epicsPrintf("acr944xAdd: cardBase %#x out of range for A16\n",
	    		cardBase);
	    return -1;
	}
	am = VME_AM_SUP_SHORT_IO;
	break;
	
    case 24:
    	if ((cardBase < 0) || (cardBase > 0xfffc00)) {
	    epicsPrintf("acr944xAdd: cardBase %#x out of range for A24\n", 
	    		cardBase);
	    return -1;
	}
	am = VME_AM_STD_SUP_DATA;
	break;
	
    default:
    	epicsPrintf("acr944xAdd: addrType %d must be 16 or 24\n", addrType);
	return -1;
    }
    
    if ((cardBase & 0x3ff) != 0) {
    	epicsPrintf("acr944xAdd: cardBase %#x not aligned correctly\n");
	return -1;
    }
    
    if ((intLevel < 0) || (intLevel > 7)) {
	epicsPrintf("acr944xAdd: intLevel %d out of range (0 - 7)\n", 
		    intLevel);
	return -1;
    }
    
    if ((intVec != 0) &&	/* Interrupts are disabled if intVec == 0 */
    	((intVec < 64) || (intVec > 255))) {
	epicsPrintf("acr944xAdd: intVec %d out of range (64 - 255)\n", intVec);
	return -1;
    }
    
    if (sysBusToLocalAdrs(am, (char *) cardBase, (char **) &pcard) == ERROR) {
	epicsPrintf("acr944xAdd: VME bus address conversion failed\n");
	return -1;
    }
    
    if (vxMemProbe((char *) pcard, READ, sizeof(dummy), (char *) &dummy) < 0) {
	epicsPrintf("acr944xAdd: Card not responding\n");
	return -1;
    }
    
    for (i=0; i < sizeof(id) - 1; i++) {
    	if ((pcard->idProm[i] & 0xff) != id[i]) {
	    epicsPrintf("acr944xAdd: Card at A%d %#x is not an ACR944x\n", 
	    		addrType, cardBase);
	return -1;
	}
    }
    
    switch (pcard->idProm[11] & 0xff) {
    case '0':
    	cardType = cardIO;
	break;
	
    case '3':
    	cardType = cardOut;
	break;
	
    case '7':
    	cardType = cardIn;
	break;
	
    default:
    	epicsPrintf("acr944xAdd: Card type ACR944%c not supported\n", 
		    addrType);
	return -1;
    }
    
    pcard->boardStatus &= ~BSR_INT_ENABLE;	/* Disable int's first */
    pcard->boardStatus |= BSR_LED_RED;		/* Turn OFF red LED */
    
    pdpvt = calloc(1, sizeof(struct dpvt));
    if (pdpvt == NULL) {
    	epicsPrintf("acr944xAdd: Out of memory\n");
	return -1;
    }
    
    if (intVec &&
    	intConnect(INUM_TO_IVEC(intVec), intRoutine, (int) pdpvt) < 0) {
    	epicsPrintf("acr944xAdd: Interrupt connect failed\n");
	return -1;
    }
    if (intLevel) sysIntEnable(intLevel);
    
    if (intVec) for (i=0; i<8; i++) pcard->intVector[i] = intVec;
    
    pcard->intEnable = 0;	/* Start with all int's off */
    pcard->intStatus = 0xff;	/* Clear any pending int's */
    pcard->intTypeSel = 0xff;	/* Int's on Change-Of-State */
    
    pdpvt->addrType = addrType;
    pdpvt->cardBase = cardBase;
    pdpvt->pcard    = pcard;
    pdpvt->cardType = cardType;
    pdpvt->intVec   = intVec;
    
    device[cardNum] = pdpvt;
    pcard->boardStatus |= BSR_INT_ENABLE;	/* Finally enable int's */
    return 0;
}


/* Status report function */

static long report(int level) {
    struct dpvt *pdpvt;
    volatile struct acr944x *pcard;
    int i;
    const static char *ctyp[3] = {"Acr9440", "Acr9443", "Acr9447"};
    
    for (i=0; i < MAX_CARDS; i++) {
    	pdpvt = device[i];
	if (pdpvt != NULL) {
	    pcard = pdpvt->pcard;
	    printf("  %d: %s at A%d %#lx (%p), vector %#x\n", i,
		    ctyp[pdpvt->cardType], pdpvt->addrType, 
		    pdpvt->cardBase, pcard, pdpvt->intVec);
	    
	    if (level) {
	    	printf("    Board Status = %#x\n", pcard->boardStatus & 0xff);
	    	printf("    Int Vector 0 = %#x\n", pcard->intVector[0] & 0xff);
	    	printf("    Int Vector 1 = %#x\n", pcard->intVector[1] & 0xff);
	    	printf("    Int Vector 2 = %#x\n", pcard->intVector[2] & 0xff);
	    	printf("    Int Vector 3 = %#x\n", pcard->intVector[3] & 0xff);
	    	printf("    Int Vector 4 = %#x\n", pcard->intVector[4] & 0xff);
	    	printf("    Int Vector 5 = %#x\n", pcard->intVector[5] & 0xff);
	    	printf("    Int Vector 6 = %#x\n", pcard->intVector[6] & 0xff);
	    	printf("    Int Vector 7 = %#x\n", pcard->intVector[7] & 0xff);
	    	printf("    Int Status   = %#x\n", pcard->intStatus & 0xff);
	    	printf("    Int Enable   = %#x\n", pcard->intEnable & 0xff);
	    	printf("    Int Polarity = %#x\n", pcard->intPolarity & 0xff);
	    	printf("    Int Type Sel = %#x\n", pcard->intTypeSel & 0xff);
	    	printf("    Int Pattern  = %#x\n", pcard->intPattern & 0xff);
	    	printf("    Input Data   = %#x\n", pcard->inputData);
	    	printf("    Output Data  = %#x\n", pcard->outputData);
	    }
	}
    }
    
    return 0;
}


/* Use a driver entry table for a device-neutral report function */

struct drvet drvAcr944x = {
    2, report, NULL
};


/*************** This is the Binary Output device support ***************/

static long initBo(struct boRecord *precord) {
    struct dpvt *pdpvt;
    
    precord->pact = 1;
    if ((precord->out.type != VME_IO) || 
	(precord->out.value.vmeio.card < 0) ||
	(precord->out.value.vmeio.card >= MAX_CARDS) ||
	(precord->out.value.vmeio.signal < 0) ||
	(precord->out.value.vmeio.signal > 15)) {
	epicsPrintf("devBoAcr944x: %s.OUT Bad hardware link address\n",
		    precord->name);
	return S_dev_badOutType;
    }
    
    pdpvt = device[precord->out.value.vmeio.card];
    if (pdpvt == NULL) {
    	epicsPrintf("devBoAcr944x: %s.OUT No card %d registered\n",
		    precord->name, precord->out.value.vmeio.card);
	return S_dev_badCard;
    }
    
    if (pdpvt->cardType == cardIn) {
    	epicsPrintf("devBoAcr944x: %s.OUT Card %d only has inputs!\n",
		    precord->name, precord->out.value.vmeio.card);
	return S_dev_badCard;
    }
    
    pdpvt->pcard->boardStatus |= BSR_LED_GREEN;	/* Flag it in use */
    
    precord->dpvt = pdpvt;
    precord->mask = 1 << precord->out.value.vmeio.signal;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


static long writeBo(struct boRecord *precord) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    int lockKey;
    
    lockKey = intLock();
    pdpvt->pcard->outputData = (pdpvt->pcard->outputData & ~precord->mask) | 
    			      (precord->rval & precord->mask);
    intUnlock(lockKey);
    
    return 0;
}


struct dsetBo {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct boRecord *precord);
    long (*getIoIntInfo)(int cmd, struct boRecord *precord, IOSCANPVT *ppvt);
    long (*write)(struct boRecord *precord);
} devBoAcr944x = {
    5, NULL, NULL, initBo, NULL, writeBo
};


/********** This is the Multi-Bit Binary Output device support **********/

static long initMbbo(struct mbboRecord *precord) {
    struct dpvt *pdpvt;
    
    precord->pact = 1;
    if ((precord->out.type != VME_IO) || 
	(precord->out.value.vmeio.card < 0) ||
	(precord->out.value.vmeio.card >= MAX_CARDS) ||
	(precord->out.value.vmeio.signal < 0) ||
	(precord->out.value.vmeio.signal + precord->nobt > 15)) {
	epicsPrintf("devMbboAcr944x: %s.OUT Bad hardware link address\n",
		    precord->name);
	return S_dev_badOutType;
    }
    
    pdpvt = device[precord->out.value.vmeio.card];
    if (pdpvt == NULL) {
    	epicsPrintf("devMbboAcr944x: %s.OUT No card %d registered\n",
		    precord->name, precord->out.value.vmeio.card);
	return S_dev_badCard;
    }
    
    if (pdpvt->cardType == cardIn) {
    	epicsPrintf("devMbboAcr944x: %s.OUT Card %d only has inputs!\n",
		    precord->name, precord->out.value.vmeio.card);
	return S_dev_badCard;
    }
    
    pdpvt->pcard->boardStatus |= BSR_LED_GREEN;	/* Flag it in use */
    
    precord->dpvt = pdpvt;
    precord->shft = precord->out.value.vmeio.signal;
    precord->mask <<= precord->shft;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


static long writeMbbo(struct mbboRecord *precord) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    int lockKey;
    
    lockKey = intLock();
    pdpvt->pcard->outputData = (pdpvt->pcard->outputData & ~precord->mask) | 
    			      (precord->rval & precord->mask);
    intUnlock(lockKey);
    
    return 0;
}


struct dsetMbbo {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct mbboRecord *precord);
    long (*getIoIntInfo)(int cmd, struct mbboRecord *precord, IOSCANPVT *ppvt);
    long (*write)(struct mbboRecord *precord);
} devMbboAcr944x = {
    5, NULL, NULL, initMbbo, NULL, writeMbbo
};


/******* This is the Multi-Bit Binary Output Direct device support *******/

static long initMbboDirect(struct mbboDirectRecord *precord) {
    struct dpvt *pdpvt;
    
    precord->pact = 1;
    if ((precord->out.type != VME_IO) || 
	(precord->out.value.vmeio.card < 0) ||
	(precord->out.value.vmeio.card >= MAX_CARDS) ||
	(precord->out.value.vmeio.signal < 0) ||
	(precord->out.value.vmeio.signal + precord->nobt > 15)) {
	epicsPrintf("devMbboDirectAcr944x: %s.OUT Bad hardware link address\n",
		    precord->name);
	return S_dev_badOutType;
    }
    
    pdpvt = device[precord->out.value.vmeio.card];
    if (pdpvt == NULL) {
    	epicsPrintf("devMbboDirectAcr944x: %s.OUT No card %d registered\n",
		    precord->name, precord->out.value.vmeio.card);
	return S_dev_badCard;
    }
    
    if (pdpvt->cardType == cardIn) {
    	epicsPrintf("devMbboDirectAcr944x: %s.OUT Card %d only has inputs!\n",
		    precord->name, precord->out.value.vmeio.card);
	return S_dev_badCard;
    }
    
    pdpvt->pcard->boardStatus |= BSR_LED_GREEN;	/* Flag it in use */
    
    precord->dpvt = pdpvt;
    precord->shft = precord->out.value.vmeio.signal;
    precord->mask <<= precord->shft;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


static long writeMbboDirect(struct mbboDirectRecord *precord) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    int lockKey;
    
    lockKey = intLock();
    pdpvt->pcard->outputData = (pdpvt->pcard->outputData & ~precord->mask) | 
    			      (precord->rval & precord->mask);
    intUnlock(lockKey);
    
    return 0;
}


struct dsetMbboDirect {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct mbboDirectRecord *precord);
    long (*getIoIntInfo)(int cmd, struct mbboDirectRecord *, IOSCANPVT *ppvt);
    long (*write)(struct mbboDirectRecord *precord);
} devMbboDirectAcr944x = {
    5, NULL, NULL, initMbboDirect, NULL, writeMbboDirect
};


/**************** This is the Binary Input device support ****************/

static long initBi(struct biRecord *precord) {
    struct dpvt *pdpvt;
    
    precord->pact = 1;
    if ((precord->inp.type != VME_IO) || 
	(precord->inp.value.vmeio.card < 0) ||
	(precord->inp.value.vmeio.card >= MAX_CARDS) ||
	(precord->inp.value.vmeio.signal < 0) ||
	(precord->inp.value.vmeio.signal > 15)) {
	epicsPrintf("devBiAcr944x: %s.INP Bad hardware link address\n",
		    precord->name);
	return S_dev_badOutType;
    }
    
    pdpvt = device[precord->inp.value.vmeio.card];
    if (pdpvt == NULL) {
    	epicsPrintf("devBiAcr944x: %s.INP No card %d registered\n",
		    precord->name, precord->inp.value.vmeio.card);
	return S_dev_badCard;
    }
    
    if (pdpvt->cardType == cardOut) {
    	epicsPrintf("devBiAcr944x: %s.INP Card %d only has outputs!\n",
		    precord->name, precord->inp.value.vmeio.card);
	return S_dev_badCard;
    }
    
    pdpvt->pcard->boardStatus |= BSR_LED_GREEN;	/* Flag it in use */
    
    precord->dpvt = pdpvt;
    precord->mask = 1 << precord->inp.value.vmeio.signal;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


static long getBiIntInfo(int cmd, struct biRecord *precord, IOSCANPVT *ppvt) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    int signal = precord->inp.value.vmeio.signal;
    
    if (signal > 7) {
    	epicsPrintf("devBiAcr944x: %s.INP Signal %d cannot interrupt\n",
		    precord->name, precord->inp.value.vmeio.signal);
	*ppvt = NULL;
    } else {
	if (pdpvt->ioscan[signal] == NULL) {
    	    scanIoInit(&pdpvt->ioscan[signal]);
	    pdpvt->pcard->intEnable |= precord->mask;
	}
	*ppvt = pdpvt->ioscan[signal];
    }
    return 0;
}


static long readBi(struct biRecord *precord) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    
    precord->rval = pdpvt->pcard->inputData & precord->mask;
    
    return 0;   /* Convert */
}


struct dsetBi {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct biRecord *precord);
    long (*getIoIntInfo)(int cmd, struct biRecord *precord, IOSCANPVT *ppvt);
    long (*read)(struct biRecord *precord);
} devBiAcr944x = {
    5, NULL, NULL, initBi, getBiIntInfo, readBi
};


/*********** This is the Multi-Bit Binary Input device support ***********/

static long initMbbi(struct mbbiRecord *precord) {
    struct dpvt *pdpvt;
    
    precord->pact = 1;
    if ((precord->inp.type != VME_IO) || 
	(precord->inp.value.vmeio.card < 0) ||
	(precord->inp.value.vmeio.card >= MAX_CARDS) ||
	(precord->inp.value.vmeio.signal < 0) ||
	(precord->inp.value.vmeio.signal + precord->nobt > 15)) {
	epicsPrintf("devMbbiAcr944x: %s.INP Bad hardware link address\n",
		    precord->name);
	return S_dev_badOutType;
    }
    
    pdpvt = device[precord->inp.value.vmeio.card];
    if (pdpvt == NULL) {
    	epicsPrintf("devMbbiAcr944x: %s.INP No card %d registered\n",
		    precord->name, precord->inp.value.vmeio.card);
	return S_dev_badCard;
    }
    
    if (pdpvt->cardType == cardOut) {
    	epicsPrintf("devMbbiAcr944x: %s.INP Card %d only has outputs!\n",
		    precord->name, precord->inp.value.vmeio.card);
	return S_dev_badCard;
    }
    
    pdpvt->pcard->boardStatus |= BSR_LED_GREEN;	/* Flag it in use */
    
    precord->dpvt = pdpvt;
    precord->shft = precord->inp.value.vmeio.signal;
    precord->mask <<= precord->shft;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


static long getMbbiIntInfo(int cmd, struct mbbiRecord *precord, IOSCANPVT *ppvt) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    
    if (precord->inp.value.vmeio.signal > 7) {
    	epicsPrintf("devMbbiAcr944x: %s.INP Signal %d cannot interrupt\n",
		    precord->name, precord->inp.value.vmeio.signal);
	*ppvt = NULL;
    } else {
	if (pdpvt->ioscan[8] == NULL) {
    	    scanIoInit(&pdpvt->ioscan[8]);
	    pdpvt->pcard->intEnable |= precord->mask;
	}
	*ppvt = pdpvt->ioscan[8];
    }
    return 0;
}


static long readMbbi(struct mbbiRecord *precord) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    
    precord->rval = pdpvt->pcard->inputData & precord->mask;
    
    return 0;   /* Convert */
}


struct dsetMbbi {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct mbbiRecord *precord);
    long (*getIoIntInfo)(int cmd, struct mbbiRecord *precord, IOSCANPVT *ppvt);
    long (*read)(struct mbbiRecord *precord);
} devMbbiAcr944x = {
    5, NULL, NULL, initMbbi, getMbbiIntInfo, readMbbi
};


/******* This is the Multi-Bit Binary Direct Input device support *******/

static long initMbbiDirect(struct mbbiDirectRecord *precord) {
    struct dpvt *pdpvt;
    
    precord->pact = 1;
    if ((precord->inp.type != VME_IO) || 
	(precord->inp.value.vmeio.card < 0) ||
	(precord->inp.value.vmeio.card >= MAX_CARDS) ||
	(precord->inp.value.vmeio.signal < 0) ||
	(precord->inp.value.vmeio.signal + precord->nobt > 15)) {
	epicsPrintf("devMbbiDirectAcr944x: %s.INP Bad hardware link address\n",
		    precord->name);
	return S_dev_badOutType;
    }
    
    pdpvt = device[precord->inp.value.vmeio.card];
    if (pdpvt == NULL) {
    	epicsPrintf("devMbbiDirectAcr944x: %s.INP No card %d registered\n",
		    precord->name, precord->inp.value.vmeio.card);
	return S_dev_badCard;
    }
    
    if (pdpvt->cardType == cardOut) {
    	epicsPrintf("devMbbiDirectAcr944x: %s.INP Card %d only has outputs!\n",
		    precord->name, precord->inp.value.vmeio.card);
	return S_dev_badCard;
    }
    
    pdpvt->pcard->boardStatus |= BSR_LED_GREEN;	/* Flag it in use */
    
    precord->dpvt = pdpvt;
    precord->shft = precord->inp.value.vmeio.signal;
    precord->mask <<= precord->shft;
    
    precord->pact = 0;	/* Initialized Ok, allow record to process */
    return 0;
}


static long getMbbiDirectIntInfo(int cmd, struct mbbiDirectRecord *precord, 
				 IOSCANPVT *ppvt) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    
    if (precord->inp.value.vmeio.signal > 7) {
    	epicsPrintf("devMbbiDirectAcr944x %s.INP Signal %d cannot interrupt\n",
		    precord->name, precord->inp.value.vmeio.signal);
	*ppvt = NULL;
    } else {
	if (pdpvt->ioscan[8] == NULL) {
    	    scanIoInit(&pdpvt->ioscan[8]);
	    pdpvt->pcard->intEnable |= precord->mask;
	}
	*ppvt = pdpvt->ioscan[8];
    }
    return 0;
}


static long readMbbiDirect(struct mbbiDirectRecord *precord) {
    struct dpvt *pdpvt = (struct dpvt *) precord->dpvt;
    
    precord->rval = pdpvt->pcard->inputData & precord->mask;
    
    return 0;   /* Convert */
}


struct dsetMbbiDirect {
    long number;
    long (*report)(int level);
    long (*initialize)(int pass);
    long (*initRecord)(struct mbbiDirectRecord *precord);
    long (*getIoIntInfo)(int cmd, struct mbbiDirectRecord *, IOSCANPVT *ppvt);
    long (*read)(struct mbbiDirectRecord *precord);
} devMbbiDirectAcr944x = {
    5, NULL, NULL, initMbbiDirect, getMbbiDirectIntInfo, readMbbiDirect
};


